from struct import unpack, pack
import sys
from fileinfile import FileInFile

def filename_id(fn):
    crc = ord(fn[0]) % 128
    for i in range(1,len(fn)):
        crc = ((crc * 0x25) + (ord(fn[i])%128)) % (2 ** 32)
    return crc & (0x400 - 1)

class Archive(object):
    def __init__(self, pcidx,  pcbin):
        a, b = unpack('<II', pcidx.read(8))
        self.tablebyids = []
        for i in range(a):
            s = pcidx.read(9)
            li = unpack('<II', s[:-1])
            self.tablebyids.append(li)
        self.extra_start = pcidx.tell()
        self.extra = pcidx.read()
        self.pcidx = pcidx
        self.pcbin = pcbin
        self.replacements = {}
        self.build_file_system()

    def add_new_replacement(self, fn, data):
        self.replacements[fn] = data
    
    def get_offseting_table(self):
        tbl = []
        l = []
        for fn in self.replacements:
            of, sz = self.get_file_info(fn)
            l += [(of, len(self.replacements[fn])-sz)]

        l.sort(lambda x, y: x[0] < y[0])
        c = 0
        for o, s in l:
            c += s
            tbl.append((o,c))

        return tbl

    def get_offset_by_table(self, no, tbl):
        r = no
        for o, s in tbl:
            if no > o:
                r = no+s
        return r

    def oldrepack(self, pcidxn, pcbinn, publish=None):
        repl = {}
        replbyo = {}
        for fn in self.replacements:
            i = filename_id(fn)
            of, sz = self.get_file_info(fn)
            repl[i] = fn
            replbyo[of] = fn

        tbl = self.get_offseting_table()
        #print tbl
        reloc = []

        s = ''
        nf = len(self.tablebyids)

        s += pack('<II', nf, 0)
        for i in range(nf):
            a, b = self.tablebyids[i]
            if a != 0 or b != 0:
                if a % 2 == 0:
                    of, sz = decode(a, b)
                    nof = self.get_offset_by_table(of, tbl)
                    nsz = sz
                    if i  in repl:
                        fn = repl[i]
                        nsz = len(self.replacements[fn])
                        reloc.append((nof,self.replacements[fn]))
                    else:
#                        print ('ID', nof, of, sz)
                        reloc.append((nof,of,sz))
                    a2, b2 = encode(nof, nsz)
                    s += pack('<II', a2, b2)[:6]+'\x42\x00\x00'
                else:
                    s += pack('<II', a, b)[:6]+'\x42\x00\x00'
            else:
                s += '\x00' * 9

        bt = self.unpack_bottom_table()

        ks = bt.keys()
        ks.sort()

        for i in ks:
            lf = bt[i]
            s += chr(len(lf))
            for i in range(len(lf)):
                of, sz, char, pos = lf[i]
                nof = self.get_offset_by_table(of, tbl) 
                if i == 0:
                    sz = lf[1][0]-of                
                nsz = sz
                if of in replbyo:
                    fn = replbyo[of]
                    nsz = len(self.replacements[fn])
                    reloc.append((nof,self.replacements[fn]))
                else:
#                    print ('BTE', nof, of, sz)
                    reloc.append((nof,of,sz))
                s += pack('<I', nof/4)
                if i != 0:
                    s += pack('<II', nsz/4, 0)
                s += char+chr(pos)+'\x00'

        d = len(s)
        od = (d + 0x3) & ~0x3
        s += '\x00' * (od-d)

        pcidxn.write(s)

        reloc.sort(lambda x, y: cmp(x[0], y[0]))

#        print 'Rebuilding pc.bin'

#        fo = file(self.pcbin_fn+'.dup', 'wb')
#        fi = file(self.pcbin_fn, 'rb')

        for i in range(len(reloc)):
            if len(reloc[i]) == 2:
                o, d = reloc[i]
                pcbinn.seek(o)
                pcbinn.write(d)
 #               print 'New file at %08x %08x' % (o, o+len(d))
                if publish != None:
                    publish.write('(%d,%d),' % (o, len(d)))
            else:
                no, o, s = reloc[i]
                if publish != None:
                    publish.write('(%d,%d,%d),' % (no,o,s))
                self.pcbin.seek(o)
                pcbinn.seek(no)
                pcbinn.write(self.pcbin.read(s))
#                if o != no:
#                    print 'Reloc from %08x to %08x %08x' % (o, no, no+s)


    def unpack_bottom_table(self):
        i = 0
        extras = {}
        while i < len(self.extra):
            si = i
            ne = ord(self.extra[i])
            if ne == 0:
                break
            i += 1
            lf = []
            for k in range(ne):
                of = unpack('<I', self.extra[i:i+4])[0]
                i += 4
                if k == 0:
                    sz = 0
                else:
                    sz = unpack('<I', self.extra[i:i+4])[0]
                    i += 8
                c = self.extra[i]
                pos = ord(self.extra[i+1])
                i += 3
                lf.append((of*4, sz*4, c, pos))
            extras[si] = lf

        return extras

    def get_file(self, id, fn=''):
        fl = self.files[id]
        if type(fl) == type([]):
            for char, pos, f in fl:
                if pos < len(fn):
                    if fn[pos] == char:
                        return f
        else:
            return fl
        return None

    def get_file_named(self, fn):
        return self.get_file(filename_id(fn),fn)

    def set_file(self, nf, id, fn=''):
        fl = self.files[id]
        if type(fl) == type([]):
            for i in range(len(fl)):
                char, pos, f = fl[i]
                if pos < len(fn):
                    if fn[pos] == char:
                        fl[i] = (char, pos, nf)
                        return
        else:
            self.files[id] = nf

    def set_file_named(self, nf, fn):
        self.set_file(nf, filename_id(fn), fn)

    def build_file_system(self):      
        self.files = []
        bte = self.unpack_bottom_table()

        for i in range(len(self.tablebyids)):
            a, b = self.tablebyids[i]
            if a == 0 and b == 0:
                self.files.append(None)
                continue
            if a % 2 == 1:
#                r2 = b
#                r6 = (a << 6) % (2 ** 32)
#                r3 = r6 >> 7 # dir table entry
#                r11 = (a >> 26) + (b % (2 ** 16))*64
#                idx = r3 - self.extra_start
                ra, rb = decode(a,b)
                r11 = rb
                idx = ra/4 - self.extra_start
                flist = []
                for of, sz, char, pos in bte[idx]:
                    if sz == 0:
                        nsz = r11
                    else:
                        nsz = sz
                    flist.append((char, pos, FileInFile(self.pcbin,of,nsz)))
                self.files.append(flist)
            else:
                of, sz = decode(a,b)
                self.files.append(FileInFile(self.pcbin, of, sz))

    def repack(self, pcidxn, pcbinn):
        def align16(d):
            return (d + 0x1f) & ~0x1f
        nfiles = len(self.files)
        pcidxn.write(pack('<II', nfiles, 0))
        cof = 0
        bteof = nfiles*9+8
        bte = ''

        for i in range(nfiles):
            fl = self.files[i]
            if fl == None:
                pcidxn.write('\x00'*9)
                continue
            if type(fl) == type([]):
                bte += chr(len(fl))
                for j in range(len(fl)):
                    char, pos, f = fl[j]
                    f.seek(0,2)
                    fsize = f.tell()
                    f.seek(0)
                    bte += pack('<I', cof/4)
                    if j == 0:
                        a, b = encode(bteof*4, fsize)
                        pcidxn.write(pack('<II', a+1, b)[:6]+'\x42\x00\x00')            
                    else:
                        bte += pack('<II', fsize/4, 0)
                    bte += char+chr(pos)+'\x00'
                    pcbinn.write(f.read())
                    nof = align16(cof+fsize)
                    pcbinn.write('\x00'*(nof-(cof+fsize)))
                    cof = nof
                bteof = len(bte)+nfiles*9+8
            else:
                fl.seek(0,2)
                fsize = fl.tell()
                fl.seek(0)
                pcbinn.write(fl.read())
                a, b = encode(cof, fsize)
                pcidxn.write(pack('<II', a, b)[:6]+'\x42\x00\x00')
                nof = align16(cof+fsize)
                pcbinn.write('\x00'*(nof-(cof+fsize)))
                cof = nof

        pcidxn.write(bte)

def decode(a, b):
    of = (a % (2 ** 26)) * 2
    sz = ((a >> 26) + (b % (2 ** 16))*64)*4
    return (of,sz)

def encode(of,sz):
    a = (of / 2) + (((sz / 4) % 64) << 26 )
    b = ((sz/4)/64)
    return (a,b)

files = {
    'system' : {
        'rom' : {
            'JD_message_rom.idx' : {},
            'JD_message.pak' : {},
            'JD_message_res.pak' : {},
            'JD_message_res_rom.idx' : {},
            'JH_freepapermes.pak' : {},
            'JH_freepapermes_rom.idx' : {},
            'JH_questtext.pak' : {},
            'JH_questtext_rom.idx' : {},
            'JH_uwasatext.pak' : {},
            'JH_uwasatext_rom.idx' : {},
            'sysdata.pak' : {},
            'sysdata_rom.idx' : {},
            'entrydata_rom.idx' : {},
            'entrydata.pak' : {},
            }
        },
    'menu' : {
        'font' : {
            'UseMoji_image.bin' : {},
            'UseMoji_table.bin' : {},
            'FontWidthTable.bin' : {},
            'fontpalette.bin' : {},
            },
        'dsl_bin' : {
            'name_dsl.bin' : {},
            },
        'atl_rom' : {
            'atl.pak' : {},
            'atl_rom.idx' : {},
            },
        'nc_rom': {
            'nc.pak' : {},
            'nc_rom.idx' : {},
            },
        },
#    'effect' : {
#        'magic' : {
#            },
#        },
    'battle' : {
        'bt_title' : {
            'Title_Quest.ebn' : {},
            'Title_Quest.obn' : {},
            },
        },
    'event' : {
        'rom' : {
            'ev_all.pak' : {},
            'ev_all_rom.idx' : {},
            'btscr_all.pak' : {},
            'btscr_all_rom.idx' : {},
            },
        },
    }

def flat_list(d=files):
    l = []
#    for i in range(300):
#        files['effect']['magic']['m%03d'%i] = {}
    for k in d:
        if d[k] == {}:
            l += [k]
        else:
            nl = flat_list(d[k])
            l += map(lambda x: k+'/'+x, nl)
    return l

def flat_list_by_ids():
    l = {}
    for fn in flat_list():
        id = filename_id(fn)
        if id in l:
            l[id] = l[id]+[fn]
        else:
            l[id] = [fn]
    return l

if __name__ == "__main__":
 #   test('system/rom/JD_message_rom.idx')
 #   test('system/rom/JD_message.pak')
 #   test('system/rom/JD_message_res.pak')
 #   test('system/rom/JD_message_res_rom.idx')
 #   test('menu/dsl_bin/name_dsl.bin')

 #   test('menu/font/FontWidthTable.bin')
 #   test('menu/font/fontpalette.bin')
   
 #   r, s = encode(0x450be98, 0x5c7e0)
 #   print map(hex, (r,s))
 #   print repr(pack('<II', r,s))
 #   print map(hex,decode(r,s))
#    print map(hex, decode(0xc231b660, 0x39ce))
  #  l = flat_list(files)
  #  l.sort()

    arcidx = Archive('ffa2/master/pc.idx', 'ffa2/master/pc.bin')
    arcidx.extract(sys.argv[1])
    
#    for f in l:
#        arcidx.extract(f, pcbin)
#    print arcidx.unpack_bottom_table()

#    f = file('test', 'wb')
 #   f.write(arcidx.repack())
 #   f.close()

 #   pcidx.close()
 #   pcbin.close()

# message_res_pak colle a message.pak ?
