import os
import sys
from tempfile import TemporaryFile
import wx
from wx import xrc
import Image
import urllib
import urllib2
#from app_xrc import get_resources, xrcpatch_dialog, xrcfetch_dialog, xrccommit_dialog, xrcsettings_dialog
import app_xrc
from nds import NDS
from archive import Archive, flat_list_by_ids
from table import hexdump_with_table, lookup, get_table, reverse
#from mestable import get_table, MessageTable
from idxandpak import IdxAndPak
from struct import pack, unpack
from stringfile import StringFile
from fileinfile import FileInFile
import database
import drawing
from graphic import Graphic
from settings import Settings
from font import FontTable, Glyph
import socket
from copy import copy, deepcopy
from preview import Preview
import zlib
import time

timeout = 30

idxpak = {
    'questtext' : ['system/rom/JH_questtext_rom.idx',
                   'system/rom/JH_questtext.pak'],
    'uwasatext' : ['system/rom/JH_uwasatext_rom.idx',
                   'system/rom/JH_uwasatext.pak'],
    'freepapermes' : ['system/rom/JH_freepapermes_rom.idx',
                   'system/rom/JH_freepapermes.pak'],
    'sysdata' : ['system/rom/sysdata_rom.idx',
                 'system/rom/sysdata.pak'],
    'entrydata' : ['system/rom/entrydata_rom.idx',
                   'system/rom/entrydata.pak'],
    'ev_all' : ['event/rom/ev_all_rom.idx',
                'event/rom/ev_all.pak'],
    'btscr_all' : ['event/rom/btscr_all_rom.idx',
                   'event/rom/btscr_all.pak'],
    'nc' : ['menu/nc_rom/nc_rom.idx', 'menu/nc_rom/nc.pak'],
    }

mestables = {
    'char names' : [0, ('sit01.png', (0,0), 0)],
    'jobs' : [1, ('sit01.png', (0,0), 0)],
    'jobs skills' : [2, ('sit01.png', (0,0), 0)],
    'item abilities' : [3, ('sit01.png', (0,0), 0)],
    'areas 1' : [4, ('sit01.png', (0,0), 0)],
    'areas 2' : [5, ('sit01.png', (0,0), 0)],
    'quest titles' : [6, ('sit01.png', (0,0), 0)],
    'guild rank' : [7, ('sit01.png', (0,0), 0)],
    'pub rumors' : [8, ('sit01.png', (0,0), 0)],
    'menu 1' : [9, ('sit01.png', (0,0), 0)],
    'free papers' : [10, ('sit01.png', (0,0), 0)],
    'equipement items' : [11, ('sit01.png', (0,0), 0)],
    'status effect' : [12, ('sit01.png', (0,0), 0)],
    'clan names' : [13, ('sit01.png', (0,0), 0)],
    'laws' : [14, ('sit01.png', (0,0), 0)],
    'clan abilities' : [15, ('sit01.png', (0,0), 0)],
    'auction' : [16, ('sit01.png', (0,0), 0)],
    'pub quest menu' : [17, ('sit01.png', (0,0), 0)],
    'equipement categories' : [18, ('sit01.png', (0,0), 0)],
    'hiragana alphabet' : [19, ('sit01.png', (0,0), 0)],
    'katakana alphabet' : [20, ('sit01.png', (0,0), 0)],
    'alphanumeric alphabet' : [21, ('sit01.png', (0,0), 0)],
    'hiragana/accents alphabet' : [22, ('sit01.png', (0,0), 0)],
    'katakana/accents alphabet' : [23, ('sit01.png', (0,0), 0)],
    'treasures' : [24,  ('sit01.png', (0,0), 0)],
    'jobs skills descr' : [30, ('sit02.png', (18,18), 16)],
    'item abi descr' : [31, ('sit02.png', (18,18), 16)],
    'areas descr 1' : [32, ('sit02.png', (18,18), 16)],
    'areas descr 2' : [33, ('sit02.png', (18,18), 16)],
    'mission descr' : [35, ('sit02.png', (18,18), 16)],
    'system messages' : [37, ('sit02.png', (18,18), 16)],
    'jobs descr' : [38, ('sit02.png', (18,18), 16)],
    'equipement descr' : [39, ('sit02.png', (18,18), 16)],
    'koed allies' : [40, ('sit02.png', (18,18), 16)],
    'battle messages' : [41, ('sit02.png', (18,18), 16)],
    'laws descr' : [42, ('sit02.png', (18,18), 16)],
    'clan abilities descr' : [43, ('sit02.png', (18,18), 16)],
    'auction 2' : [44, ('sit02.png', (18,18), 16)],
    'pub quest menu 2' : [45, ('sit02.png', (18,18), 16)],
    'skill effects' : [47, ('sit02.png', (18,18), 16)],
    'leaving allies' : [48, ('sit02.png', (18,18), 16)],
    'quest result' : [49, ('sit02.png', (18,18), 16)],
    'shop stuff' : [50, ('sit02.png', (18,18), 16)],
    'boost message' : [51, ('sit02.png', (18,18), 16)],
    'system messages 2' : [52, ('sit02.png', (18,18), 16)],
    'bargain descr' : [53, ('sit02.png', (18,18), 16)],
    'auction 3' : [54, ('sit02.png', (18,18), 16)],
    'quest text' : [60],
    'free papers text' : [61],
    'rumors text' : [62],
#    'ev all' : [70],
    }

big_text_tables = { 
    60 : 'questtext',
    61 : 'freepapermes',
    62 : 'uwasatext',
    70 : 'ev_all',
    }

version='0.6'
version_type='beta'
author='deufeufeu'
date='2007-11-23'
website='http://deufeufeu.free.fr/wiki'
protocol_version='2'

def get_text_list(f, base_off=0, extra=False):
    ntxt = unpack('<H', f.read(2))[0]

    txto = []
    for j in range(ntxt):
        if extra:
            txto.append(unpack('<IIH', f.read(10)))
        else:
            txto.append(unpack('<II', f.read(8)))
    textl = []
    for j in range(ntxt):
        sz, of = txto[j][:2]
        f.seek(base_off+of)
        if extra:
            textl.append((f.read(sz), txto[j][2]))
        else:
            f.read(2)
            textl.append(f.read(sz))
    return textl

def get_story_pack(archive, font_width_table, save_table=False):
    pakname = 'ev_all'
    pak = IdxAndPak(archive.get_file_named(idxpak[pakname][0]),
                    archive.get_file_named(idxpak[pakname][1]))

    for i in range(pak.nfiles()):
        f = pak.get_file(i)
        if f == None: continue
        f.seek(0,2)
        fsize = f.tell()
        f.seek(0)
        if fsize < 16: continue
        text_offset, kanji_offset, unk1_offset, unk2_offset = \
            unpack('<IIII', f.read(16))

        if text_offset == kanji_offset:
            continue

        f.seek(text_offset)

        if save_table:
            tb = database.Table(['id', 'original', 'repl', 'repl_time', 'server', 'up', 'uptime'])
            text_list = get_text_list(f, base_off=text_offset, extra=True)
            j = 0
            for t, s in text_list:
                tb.add_row([j, t, None, '', None, '', ''])
                j += 1                        
                # hex(s), lookup(t)
            f = file('tables/table_text_%04d.csv'  % i, 'w')
            tb.save(f)
            f.close()

def get_big_text_pack(pak, font_width_table):
#    pak = IdxAndPak(archive.get_file_named(idxpak[pakname][0]),
#                    archive.get_file_named(idxpak[pakname][1]))

    bigtext = []
    for i in range(pak.nfiles()):
        f = pak.get_file(i)
        if f == None:
            bigtext.append((None,None))
            continue
        font_table_off = unpack('<H', f.read(2))[0]
        font_table_off += 2
        textl = get_text_list(f)
        text = '\x00\x00'.join(textl)
        f.seek(0, 2)
        size = f.tell()
        if font_table_off == 0:
            extra_font_table = None
        else:
            f.seek(font_table_off)
            fontf = FileInFile(f, font_table_off, size-font_table_off)
            extra_font_table = FontTable(fontf, font_width_table)

        bigtext.append((text, extra_font_table))
    return bigtext

def create_table_from_text_list(table_file, text_list):
    t = database.Table(['id', 'original', 'repl', 'repl_time', 'server', 'server_uploader', 'server_time'])
    for i in range(len(text_list)):
        if text_list[i] is not None:
            t.add_row([i, text_list[i], None, '', None, '', ''])
    t.save(table_file)
    
def stripped(original):
    stripped = ''
    for i in range(len(original)/2):
        if (ord(original[2*i+1]) >> 4 == 2) or (ord(original[2*i+1]) >> 4 == 8):
            pass
        else:
            stripped += original[2*i:2*i+2]
    return stripped

def get_server_time():
    st = list(time.gmtime())
    st[3] = st[3]+1
    return time.strftime('%Y-%d-%m %H:%M:%S', st)

class MainFrame(app_xrc.xrcMAINFRAME):
    def __init__(self, parent_app):
        app_xrc.xrcMAINFRAME.__init__(self, None)
        self.parent_app = parent_app
        self.init_frame()

    def init_frame(self):
        # Bind all events and init some controls
        self.Bind(wx.EVT_CLOSE, self.parent_app.on_quit)
        self.archivefileslist.Bind(wx.EVT_LISTBOX, self.parent_app.on_archive_file_select)
        self.idxpak_list.InsertItems(idxpak.keys(),0)
        self.idxpak_list.Bind(wx.EVT_LISTBOX, self.parent_app.on_idxpak_select)
        self.ipak = None
        self.file_in_idxpak_list.Bind(wx.EVT_LISTBOX,
                                      self.parent_app.on_fileinidx_select)
        self.local_text_mestable.Bind(wx.EVT_TEXT, self.parent_app.on_text_mestable_update)
        l = []
        self.mestables = {}
        for k in mestables:
            l.append((k,mestables[k][0]))
            self.mestables[mestables[k][0]] = k
        l.sort(lambda x,y:cmp(x[1],y[1]))

        self.mestable_list.InsertItems([e[0] for e in l] , 0)
        self.mestable_list.Bind(wx.EVT_LISTBOX, self.parent_app.on_mestable_select)
        self.copy_server_btn.Bind(wx.EVT_BUTTON, self.parent_app.on_copy_server_btn)
        self.init_special_btn.Bind(wx.EVT_BUTTON, self.parent_app.on_init_special_btn)
        self.remove_trans_btn.Bind(wx.EVT_BUTTON, self.parent_app.on_remove_trans_btn)
        self.history_btn.Bind(wx.EVT_BUTTON, self.parent_app.on_history_btn)
        self.string_mestable_list.Bind(wx.EVT_LISTBOX,
                                       self.parent_app.on_stringinmestable_select)



        self.preview_window = wx.ScrolledWindow(self, style=wx.HSCROLL|wx.VSCROLL)
        self.preview_window.SetBackgroundColour('GRAY')
        self.preview_window.SetScrollRate(20,20)
        self.parent_app.res.AttachUnknownControl('preview_window', self.preview_window)
        self.preview = Preview(self)

        self.font_tree.Bind(wx.EVT_TREE_SEL_CHANGED, self.parent_app.font_tree_select)
        self.font_display.Bind(wx.EVT_PAINT, self.parent_app.font_display_paint)
        self.data_to_display = None
        self.imgview_idxpak.Bind(wx.EVT_PAINT, self.parent_app.imgview_idxpak_paint)
        self.toolbar.Bind(wx.EVT_TOOL, self.parent_app.on_tool)

class MainApp(wx.App):
    def __getattribute__(self, k):
        try:
            return wx.App.__getattribute__(self, k)
        except AttributeError:
            return getattr(self.frame, k)

    def on_tool(self, evt):
        if evt.Id == xrc.XRCID('tool_save_db'):
            self.on_save()
        elif evt.Id == xrc.XRCID('tool_reload_db'):
            self.load_db(debug=self.debug)
        elif evt.Id == xrc.XRCID('tool_patch'):
            self.do_patch()
        elif evt.Id == xrc.XRCID('tool_fetch_db'):
            self.do_fetch_db()
        elif evt.Id == xrc.XRCID('tool_commit_db'):
            self.do_commit_db()
        elif evt.Id == xrc.XRCID('tool_settings'):
            self.do_settings()
        elif evt.Id == xrc.XRCID('tool_search'):
            self.do_search()

    def bind_dialog_buttons(self, dialog):
        dialog.ok_btn.Bind(wx.EVT_BUTTON, lambda evt: dialog.EndModal(0))
        dialog.cancel_btn.Bind(wx.EVT_BUTTON, lambda evt: dialog.EndModal(1))

    def do_search(self):
        search_dialog = app_xrc.xrcsearch_dialog(self.frame)
        self.bind_dialog_buttons(search_dialog)
        ret = search_dialog.ShowModal()

        if ret == 1:
            return

        search_text = search_dialog.search_text.GetValue().encode('utf-8')
        search_options = search_dialog.search_tables.GetSelection()

        search_dialog.Destroy()

        results = self.search(search_text, search_options)
        
        search_results_dialog = app_xrc.xrcsearch_results_dialog(self.frame)
        results_list = search_results_dialog.results_list

        formatted_results = []
        for res in results:
            tablen, strid, original, repl, server = res
            s = tablen+':'+str(strid)+':'+lookup(original)+':'
            if repl is not None:
                s += lookup(repl)+':'
            else:
                s += 'None:'
            if server is not None:
                s += lookup(server)
            else:
                s += 'None'

            formatted_results.append(s.decode('utf-8'))

        results_list.InsertItems(formatted_results, 0)
        search_results_dialog.Show()
        
    def search(self, search_text, search_options):
        enctext = reverse(search_text)
        results = []
        bsearchfield = []

        if search_options == 0:
            bsearchfield = ['original']
        elif search_options == 1:
            bsearchfield = ['repl']
        elif search_options == 2:
            bsearchfield = ['server']
        elif search_options == 3:
            bsearchfield = ['original', 'repl', 'server']

        for tablen in mestables:
            t = self.local_db.get_table(tablen)
            for i in range(t.nrows()):
                row = t.get_row(i)
                bgood = False
                for field in bsearchfield:
                    if row[field] is not None:
                        if enctext in row[field]:
                            bgood = True
                            break
                if bgood:
                    results.append((tablen, row['id'], 
                                    row['original'], row['repl'], row['server']))

        return results

    def do_settings(self):
        settings_dialog = app_xrc.xrcsettings_dialog(self.frame)

        if self.settings.original_rom is not None: settings_dialog.original_rom_path.SetValue(self.settings.original_rom)
        if self.settings.patched_rom is not None: settings_dialog.patched_rom_path.SetValue(self.settings.patched_rom)
        if self.settings.server_url is not None: settings_dialog.server_url.SetValue(self.settings.server_url)
        if self.settings.server_user is not None: settings_dialog.server_user.SetValue(self.settings.server_user)
        if self.settings.server_password is not None: settings_dialog.server_password.SetValue(self.settings.server_password)
        settings_dialog.save_btn.Bind(wx.EVT_BUTTON, lambda evt: settings_dialog.EndModal(0))
        settings_dialog.cancel_btn.Bind(wx.EVT_BUTTON, lambda evt: settings_dialog.EndModal(1))

        ret = settings_dialog.ShowModal()

        if ret == 1:
            return

        self.settings.original_rom = settings_dialog.original_rom_path.GetValue().encode('utf-8')
        self.settings.patched_rom = settings_dialog.patched_rom_path.GetValue().encode('utf-8')
        self.settings.server_url = settings_dialog.server_url.GetValue().encode('utf-8')
        self.settings.server_user = settings_dialog.server_user.GetValue().encode('utf-8')
        self.settings.server_password = settings_dialog.server_password.GetValue().encode('utf-8')

    def fetch_table(self, k, copylocal = False, debug=None):
        if debug == None:
            debug = self.frame.debug
        self.local_db_changed = True
        tableid = mestables[k][0]
        values = {'tableid' : str(tableid)}

        last_time = ''

        t = self.local_db.get_table(k)
        for i in range(t.nrows()):
            last_time = max(last_time, t.get_field(i, 'server_time'))
        values['last_time'] = last_time

#        if self.settings.server_last_update != None:
#            values['last_time'] = self.settings.server_last_update
#        else:
#            values['last_time'] = ''

        data = urllib.urlencode(values)
        req = urllib2.Request(self.settings.server_url+'get_table.php',data)
        req.add_header('Accept-encoding', 'gzip') 

        try:
            resp = urllib2.build_opener().open(req)
            data = resp.read()
#            print data
            lines = data.split('\n')
        except socket.error:
            errno, errstr = sys.exc_info()[:2]
            if errno == socket.timeout:
                pass
            return

        maxt = ''
        changed = []
        for line in lines:
            r = eval('['+line+']')

            if len(r) < 2:
                break
            changed.append(r[0])
            t.update_field(t.select_field('id', r[0])[0], 'server', r[1])
            if copylocal:
                t.update_field(t.select_field('id', r[0])[0], 'repl', r[1])
            t.update_field(t.select_field('id', r[0])[0], 'server_uploader', r[2])
            t.update_field(t.select_field('id', r[0])[0], 'server_time', r[3])
            debug.WriteText('Fetched update to table '+k+' id '+str(r[0])\
                                +' made by '+r[2]+' on '+r[3]+'\n')
            debug.Update()
            if r[3] > maxt:
                maxt = r[3]

        return maxt, changed

    def commit_table(self, k, changed):
        tableid = mestables[k][0]
        t = self.local_db.get_table(k)
        values = {'user' : self.settings.server_user, 'pass' : self.settings.server_password,
                  'protocol_version' : protocol_version }

        for i in range(t.nrows()):
            row = t.get_row(i)
            if row['repl'] != None and row['repl'] != row['server']:
                if row['repl_time'] < row['server_time']:
                    self.frame.debug.WriteText('Not commiting (%s,%d) because server version is newer.\n.' % (k, row['id']))
                elif row['id'] in changed:
                    self.frame.debug.WriteText('Not commiting (%s,%d) because server version was recently changed.\n.' % (k, row['id']))
                else:
                    values['req%d' % i] = str(tableid)+'#'+str(row['id'])+'#'+row['repl']
#                row['server'] = row['repl']

        #print repr(values)
        data = urllib.urlencode(values)

        req = urllib2.Request(self.settings.server_url+'raw_insert.php',data)
        resp = urllib2.urlopen(req)
        self.frame.debug.WriteText(resp.read()+'\n')

    def select_all_checklist(self, checklist):
        for i in range(checklist.GetCount()):
            checklist.Check(i)

    def deselect_all_checklist(self, checklist):
        for i in range(checklist.GetCount()):
            checklist.Check(i, False)

    def get_encoded_string(self, t, i, doubleg):
        row = t.get_row(t.select_field('id', i)[0])
        sizeoriginal = len(row['original'])
        if row['repl'] == None:
#            origs = row['original']
#            s = stripped(row['original'])
            origs = row['original']
            dot = reverse('.')
            s = ''
            for j in range(len(origs)/2):
                v = unpack('<H', origs[2*j:2*j+2])[0]
                if (v >> 12) == 2 or (v >> 12) == 8:
                    s += dot
                else:
                    s += origs[2*j:2*j+2]

            # Stripping 0x8000 characters
#            for j in range(len(origs)/2):
#                if origs[2*j+1] == '\x80':
#                    s += ''#'\x00\x20'
#                else:
#                    s += origs[2*j:2*j+2]
            sizecomp = len(s)
            sizeuncomp = len(s)
        else:
            os = row['repl']
            s = ''
#             best_compression = False
#             if best_compression:
#                 def compress(os):
#                     if os == '' or len(os) == 2:
#                         return os
#                     else:
#                         a, b = unpack('<HH', os[:4])
#                         if (a,b) in doubleg:
#                             c1 = compress(os[4:])
#                             c2 = compress(os[2:])
#                             if len(c1) < len(c2):
#                                 return pack('<H', doubleg[(a,b)]+0x2000)+c1
#                             else:
#                                 return os[:2]+c2
#                         else:
#                             return os[:2]+compress(os[2:])
#                 s = compress(os)
#             else:
            j = 0
            while j < len(os)/2:
                if j == len(os)/2-1:
                    s += os[2*j:2*j+2]
                    j += 1
                else:
                    a, b = unpack('<HH', os[2*j:2*j+4])
                    if (a, b) in doubleg:
                        s += pack('<H', doubleg[(a,b)]+0x2000)
                        j += 2
                    else:
                        s += os[2*j:2*j+2]
                        j += 1
            sizeuncomp = len(os)
            sizecomp = len(s)
        return s, (sizeoriginal, sizecomp, sizeuncomp)


    def do_fetch_db(self):
        fetch_dialog = app_xrc.xrcfetch_dialog(self.frame)
        fetch_dialog.fetch_button.Bind(wx.EVT_BUTTON, lambda evt: fetch_dialog.EndModal(0))
        fetch_dialog.cancel_button.Bind(wx.EVT_BUTTON, lambda evt: fetch_dialog.EndModal(1))

        fetch_dialog.select_all.Bind(wx.EVT_BUTTON, lambda evt: self.select_all_checklist(fetch_dialog.tables_list))
        fetch_dialog.deselect_all.Bind(wx.EVT_BUTTON, lambda evt: self.deselect_all_checklist(fetch_dialog.tables_list))

        l = []
        for k in mestables:
            l.append((k,mestables[k][0]))
        l.sort(lambda x,y:cmp(x[1],y[1]))

        fetch_dialog.tables_list.InsertItems([e[0] for e in l],0)

        ret = fetch_dialog.ShowModal()

        if ret == 1:
            return        

        dlg = wx.ProgressDialog("Fetching tables", "Starting the fetching process",
                                maximum=len(l), style=wx.PD_APP_MODAL | wx.PD_ELAPSED_TIME)
        tbcount = 0
        maxt = ''
        for i in range(len(l)):
            if fetch_dialog.tables_list.IsChecked(i):
                dlg.Update(i, 'Fetching table %s from server\n' % l[i][0])
#                self.debug.WriteText('Fetching table %s from server\n' % l[i][0]) 
                rt,changed = self.fetch_table(l[i][0], copylocal=fetch_dialog.copy_local.IsChecked())
                if rt > maxt: maxt = rt
                tbcount += 1

        if fetch_dialog.copy_local.IsChecked() and tbcount == len(l):
            if self.settings.server_last_update is None:
                self.settings.server_last_update = maxt
            elif maxt > self.settings.server_last_update:
                self.settings.server_last_update = maxt
            self.on_save()
        dlg.Destroy()

    def do_commit_db(self):
        values = {'user' : self.settings.server_user, 'pass' : self.settings.server_password }
        values['locking'] = 0
        data = urllib.urlencode(values)
        req = urllib2.Request(self.settings.server_url+'lock.php',data)
        resp = urllib2.urlopen(req)

        t = resp.read()

        if 'locked' in t:
            dlg = wx.MessageDialog(self.frame, 'DATABASE ERROR : %s' % t,
                                   'Database Error', wx.ICON_ERROR | wx.OK)
            dlg.ShowModal()
            self.frame.debug.WriteText(t+'\n')
            return

        commit_dialog = app_xrc.xrccommit_dialog(self.frame)
        commit_dialog.commit_button.Bind(wx.EVT_BUTTON, lambda evt: commit_dialog.EndModal(0))
        commit_dialog.cancel_button.Bind(wx.EVT_BUTTON, lambda evt: commit_dialog.EndModal(1))

        commit_dialog.select_all.Bind(wx.EVT_BUTTON, lambda evt: self.select_all_checklist(commit_dialog.tables_list))
        commit_dialog.deselect_all.Bind(wx.EVT_BUTTON, lambda evt: self.deselect_all_checklist(commit_dialog.tables_list))

        l = []
        for k in mestables:
            l.append((k,mestables[k][0]))
        l.sort(lambda x,y:cmp(x[1],y[1]))

        commit_dialog.tables_list.InsertItems([e[0] for e in l],0)

        ret = commit_dialog.ShowModal()

        if ret == 0:
            dlg = wx.ProgressDialog("Committing tables", "Starting the committing process",
                                    maximum=len(l), style=wx.PD_APP_MODAL | wx.PD_ELAPSED_TIME)
            for i in range(len(l)):
                if commit_dialog.tables_list.IsChecked(i):
                    dlg.Update(i, 'Commit table %s to server\n' % l[i][0])
                    maxt, changed = self.fetch_table(l[i][0])
                    self.commit_table(l[i][0], changed)
                    self.fetch_table(l[i][0])
            dlg.Destroy()

        values['locking'] = 1
        data = urllib.urlencode(values)
        req = urllib2.Request(self.settings.server_url+'lock.php',data)
        resp = urllib2.urlopen(req)

    def do_patch(self):
        patch_dialog = app_xrc.xrcpatch_dialog(self.frame)
        patch_dialog.patch_button.Bind(wx.EVT_BUTTON, lambda evt: patch_dialog.EndModal(0))
        patch_dialog.cancel_button.Bind(wx.EVT_BUTTON, lambda evt: patch_dialog.EndModal(1))
        
        patch_dialog.select_all.Bind(wx.EVT_BUTTON, lambda evt: self.select_all_checklist(patch_dialog.tables_list))
        patch_dialog.deselect_all.Bind(wx.EVT_BUTTON, lambda evt: self.deselect_all_checklist(patch_dialog.tables_list))

        l = []
        for k in mestables:
            l.append((k,mestables[k][0]))
        l.sort(lambda x,y:cmp(x[1],y[1]))

        patch_dialog.tables_list.InsertItems([e[0] for e in l],0)

#        for i in self.settings.tables_to_patch:
        for i in range(len(l)):
            k, id = l[i]
            if id in self.settings.tables_to_patch:
                patch_dialog.tables_list.Check(i)

        ret = patch_dialog.ShowModal()

        if ret == 1:
            return

        tables_to_patch = []

        nt = []
        for i in range(len(l)):
            t = l[i][0]
            if patch_dialog.tables_list.IsChecked(i):
                nt.append(l[i][1])
                tables_to_patch.append(patch_dialog.tables_list.GetString(i))

        self.settings.tables_to_patch = nt

        self.frame.debug.WriteText('Patching...Building new files...')

        dlg = wx.ProgressDialog("Patching rom",
                                "Starting the patching process",
                                maximum=len(tables_to_patch)+4,
                                style=wx.PD_APP_MODAL | wx.PD_ELAPSED_TIME)

        count = 0

        doubles = {}

        for tablen in tables_to_patch:
            tableid = mestables[tablen][0]
            t = self.local_db.get_table(tablen)
            nentries = t.nrows()
            for i in range(nentries):
                row = t.get_row(t.select_field('id', i)[0])
                if row['repl'] != None:
                    s = row['repl']
                    if len(s) == 2:
                        continue
                    for j in range(len(s)/2-1):
                        a, b = unpack('<HH', s[2*j:2*j+4])
                        if (a >> 12) == 2 and (b >> 12) == 2:
                            if (a,b) in doubles:
                                doubles[(a,b)] += 1
                            else:
                                wa = self.font_table.glyphs[a-0x2000].width
                                wb = self.font_table.glyphs[b-0x2000].width
                                if wa + wb < 16:
                                    doubles[(a,b)] = 1

        v = [(k, doubles[k]) for k in doubles]
        v.sort(lambda x, y: cmp(x[1],y[1]))
        v.reverse()

        doublel = v
        doubleg = {}

        dlg.Update(count, 'Rebuilding the font')
        count += 1

        font_image = StringFile()

        tokeep = [0,2,313,314,315,317,376,457,480,660,665,994,995]+\
            range(161,229)+range(732,742)
        doubleg, new_font_width_table = self.font_table.serialize(font_image,
                                                                  doublel, tokeep)
        
        font_image.seek(0)
        self.archive.set_file_named(font_image, 'menu/font/UseMoji_image.bin')

        new_font_width_table += [6]*(len(self.font_width_table)-len(new_font_width_table))
        font_width = StringFile()
        font_width.write(pack('<H', len(new_font_width_table)))
        for w in new_font_width_table:
            font_width.write(chr(w))
        font_width.seek(0)

        self.archive.set_file_named(font_width, 'menu/font/FontWidthTable.bin')

        #print tables_to_patch

        for tablen in tables_to_patch:
            tableid = mestables[tablen][0]
            dlg.Update(count, 'Building table %s' % tablen)
            t = self.local_db.get_table(tablen)
            nentries = t.nrows()
            count += 1

            if tableid in big_text_tables:
                pakname = big_text_tables[tableid]
                #print tablen, pakname

                pak = IdxAndPak(
                    self.archive.get_file_named(idxpak[pakname][0]),
                    self.archive.get_file_named(idxpak[pakname][1])
                    )

#                bigtext = get_big_text_pack(pak, self.font_width_table)

                sizeuncomp = 0
                sizecomp = 0
                sizeoriginal = 0
                for i in range(nentries):
                    f = StringFile()
                    s, sized = self.get_encoded_string(t, i, doubleg)
                    sizeoriginal += sized[0]
                    sizecomp += sized[1]
                    sizeuncomp += sized[2]
                    textid = t.data[i][t.fields.index('id')]


                    nlines = 1
                    j = 0
                    start = 0
                    rs = ''
                    while j < len(s)/2:
                        if s[2*j:2*j+2] == '\x00\x60' or s[2*j:2*j+2] == '\x01\x60':
                            if (j-start)/2 > 150:
                                rs += s[2*start:2*(start+20)]+'\x00\x60'
                                rs += s[2*(start+20):2*j+2]

                            else:
                                rs += s[2*start:2*j+2]
                            start = j+1
                        j += 1

                    for j in range(len(rs)/2):
                        if rs[2*j:2*j+2] == '\x00\x60':
                            nlines += 1
                    
                    f.write(pack('<HHII', 0, 1, len(rs), 12))
                    f.write(pack('<H', nlines))
                    f.write(rs)
                    fontoff = f.tell()-2

 #                   extra_font_table = bigtext[textid][1]
 #                   extra_font_table.serialize(f)
 #                   f.seek(0)
 #                   f.write(pack('<H', fontoff))
                    f.seek(0)

                    pak.set_file(textid, f)

                idxs, paks = pak.repack()
                idx = StringFile(idxs)
                pak = StringFile(paks)
                self.archive.set_file_named(idx, idxpak[pakname][0])
                self.archive.set_file_named(pak, idxpak[pakname][1])
                self.frame.debug.WriteText('table %s was compressed from %d bytes to %d bytes (original %d)\n' % (tablen,sizeuncomp,sizecomp,sizeoriginal))
            else:
                sizeuncomp = 0
                sizecomp = 0
                sizeoriginal = 0
                f = StringFile()
                f.write(pack('<H', nentries))
                of = 2+nentries*10
                st = []

                for i in range(nentries):
                    s, sized = self.get_encoded_string(t, i, doubleg)
                    sizeoriginal += sized[0]
                    sizecomp += sized[1]
                    sizeuncomp += sized[2]
                    bres = False
                    boxsize = 1
                    for j in range(len(s)/2):
                        if s[2*j:2*j+2] == '\x00\x60' or s[2*j:2*j+2] == '\x27\x60':
                            boxsize += 1
                        if ord(s[2*j+1]) >> 4 == 8:
                            bres = True
                    if bres: boxsize += 1 << 8
                    f.write(pack('<IIH', len(s), of, boxsize))
                    st.append(s)
                    of += len(s)

                self.frame.debug.WriteText('table %s was compressed from %d bytes to %d bytes (original %d)\n' % (tablen,sizeuncomp,sizecomp,sizeoriginal))
                for s in st:
                    f.write(s)
                f.seek(0)

                self.mespak.set_file(tableid, f)

        idxs, paks = self.mespak.repack()
        idx = StringFile(idxs)
        pak = StringFile(paks)
        self.archive.set_file_named(idx, 'system/rom/JD_message_rom.idx')
        self.archive.set_file_named(pak, 'system/rom/JD_message.pak')
        

        dlg.Update(count, 'Repacking pc archive')
        count += 1
        self.frame.debug.WriteText('Repacking pc archive...')
        pcidx = TemporaryFile()
        pcbin = TemporaryFile()
        self.archive.repack(pcidx, pcbin)
        dlg.Update(count, 'Saving the new rom')
        count += 1
        self.frame.debug.WriteText('Saving the new rom...')
        pcidx.seek(0)
        pcbin.seek(0)
        self.nds.replace_file('master/pc.idx', pcidx)
        self.nds.replace_file('master/pc.bin', pcbin)


        # NO NEED FOR PATCHING NOW THAT I CAN COMPRESS
        ov3_off = self.nds.overlay_table[3][1]
        arm9_off = 0x2000000

        patch = [
#            (0x2082bec, 'arm9', arm9_off, 6, 'B'),
#            (0x213dd70, 3, ov3_off, 6, 'B'),   
#            (0x213dd88, 3, ov3_off, 6, 'B'),  
            ]

        for o, sf, so, v, t in patch:
            vs = pack('<'+t, v)
            vn = len(vs)
            ro = o-so
            if sf == 'arm9':
                oldv = unpack('<'+t, self.nds.arm9[ro:ro+vn])[0]
                #print 'Patching arm9 binary %x -> %x @%08x' % (oldv, v, o)
                self.nds.arm9 = self.nds.arm9[:ro]+vs+self.nds.arm9[ro+vn:]
            else:
                oidx = sf
                ofdata = self.nds.fat.overlays[oidx].f.read()
                ofdata = ofdata[:ro]+vs+ofdata[ro+vn:]
                self.nds.fat.overlays[oidx].f = StringFile(ofdata)

        self.nds.save(self.settings.patched_rom)

        pcidx.close()
        pcbin.close()

        dlg.Destroy()
        self.frame.debug.WriteText('done.\n')

    def on_save(self, evt=None):
        if not self.local_db_changed:
            return
        self.frame.debug.WriteText('Saving local db...') 
        for tablen in mestables:
            t = self.local_db.get_table(tablen)
            id = mestables[tablen][0]
            f = file(self.settings.tables_dir+os.sep+'table%04d.csv' % id, 'w')
            t.save(f)
            f.close()
        self.frame.debug.WriteText('done.\n')
        self.local_db_changed = False

    def need_update(self):
        values = {'time' : self.settings.server_last_update }

        data = urllib.urlencode(values)
        req = urllib2.Request(self.settings.server_url+'update_needed.php',data)

        try:
            resp = urllib2.build_opener().open(req)
            data = resp.read()
            if 'True' in data:
                return True
            else:
                return False
        except socket.error:
            errno, errstr = sys.exc_info()[:2]
            if errno == socket.timeout:
                pass
            return False

    def on_history_btn(self, evt):
        sel = self.mestable_list.GetStringSelection().replace('*','')
        id = int(self.string_mestable_list.GetStringSelection().replace('*','').split(',')[0])

        history_dlg = app_xrc.xrchistory_dialog(self.frame)

        values = {'tableid' : str(mestables[sel][0]), 'strid' : id}

        data = urllib.urlencode(values)
        req = urllib2.Request(self.settings.server_url+'history.php',data)
        req.add_header('Accept-encoding', 'gzip') 

        try:
            resp = urllib2.build_opener().open(req)
            data = resp.read()
#            print data
            lines = data.split('\n')
        except socket.error:
            errno, errstr = sys.exc_info()[:2]
            if errno == socket.timeout:
                pass
            return

        history = []
        history_dlg.choice.Clear()

        for line in lines:
            r = eval('['+line+']')

            if len(r) < 2:
                break

            strid = r[0]
            txt = lookup(r[1]).replace('[carriage_return]', '\n')
            uploader = r[2]
            time = r[3]

            history_dlg.choice.Insert(uploader+' '+time, len(history))
            history.append(txt)

        if history == []: return
            
        history_dlg.choice.Bind(wx.EVT_CHOICE,
                                lambda evt: history_dlg.text.SetValue(history[history_dlg.choice.GetSelection()].decode('utf-8')))

        history_dlg.choice.SetSelection(0)
        history_dlg.text.SetValue(history[0].decode('utf-8'))

        history_dlg.Show()

    def on_init_special_btn(self, evt):
        sel = self.mestable_list.GetStringSelection().replace('*','')
        id = int(self.string_mestable_list.GetStringSelection().replace('*','').split(',')[0])
        t = self.local_db.get_table(sel)
        row = t.get_row(t.select_field('id', id)[0])
        original = row['original']
        s = stripped(original)
        text = lookup(s).replace('[carriage_return]', '\n')
        t.update_field(t.select_field('id', id)[0], 'repl', s)
        self.local_text_mestable.ChangeValue(text.decode('utf-8'))
        self.preview.refresh()

    def on_copy_server_btn(self, evt):
        sel = self.mestable_list.GetStringSelection().replace('*','')
        id = int(self.string_mestable_list.GetStringSelection().replace('*','').split(',')[0])
        t = self.local_db.get_table(sel)
        row = t.get_row(t.select_field('id', id)[0])
        text = lookup(row['server']).replace('[carriage_return]', '\n')
        self.local_text_mestable.ChangeValue(text.decode('utf-8'))
        t.update_field(t.select_field('id', id)[0], 'repl', row['server'])
        self.preview.refresh()

    def on_remove_trans_btn(self, evt):
        sel = self.mestable_list.GetStringSelection().replace('*','')
        id = int(self.string_mestable_list.GetStringSelection().replace('*','').split(',')[0])
        t = self.local_db.get_table(sel)
        t.update_field(t.select_field('id', id)[0], 'repl', None)
        self.local_text_mestable.ChangeValue('')
        self.preview.refresh()

    def on_text_mestable_update(self, evt):
        sel = self.mestable_list.GetStringSelection().replace('*','')
        id = int(self.string_mestable_list.GetStringSelection().replace('*','').split(',')[0])
        t = self.local_db.get_table(sel)
        text = self.local_text_mestable.GetValue().encode('utf-8')
        if text == '':
            return

        enc = reverse(text.replace('\n', '[carriage_return]'))

        if enc == t.get_field(t.select_field('id',id)[0], 'repl'):
            return


        self.local_db_changed = True

        self.mestable_list.SetString(self.mestable_list.GetSelection(), '*'+sel)
        self.string_mestable_list.SetString(self.string_mestable_list.GetSelection(),
                                            '*'+str(id)+','.join(self.string_mestable_list.GetStringSelection().replace('*','').split(',')[1:]))




        t.update_field(t.select_field('id', id)[0], 'repl', enc)
        t.update_field(t.select_field('id', id)[0], 'repl_time', get_server_time())

        if enc == t.get_field(t.select_field('id',id)[0], 'server'):
            self.string_mestable_list.SetString(self.string_mestable_list.GetSelection(),
                                                str(id)+','.join(self.string_mestable_list.GetStringSelection().replace('*','').split(',')[1:]))

        self.preview.refresh()

    def font_tree_select(self, evt):
        self.font_display.Refresh()
        self.font_display.Update()

    def imgview_idxpak_paint(self, evt):
        panel = self.imgview_idxpak
        dc = wx.PaintDC(panel)
        if self.data_to_display == None:
            return

        sim = ''
        bpp = 1
        for i in range(int(len(self.data_to_display)/bpp)):
            sim += self.data_to_display[i]*3

        w = 256
        h = len(sim)/(w*3)

        bmp = wx.BitmapFromBuffer(w, h, sim[:w*h*3])
        dc.DrawBitmap(bmp, 0, 0)


    def font_display_paint(self, evt):
        panel = evt.GetEventObject()
        dc = wx.PaintDC(panel)
        sel = self.font_tree.GetSelection()
        if not sel.IsOk():
            return
        parsel = self.font_tree.GetItemParent(sel)
        if not parsel.IsOk():
            return
        glyph_table = self.font_tree.GetItemText(parsel)
        t = self.font_tree.GetItemText(sel)
#        self.frame.debug.WriteText(t+'\n')

        nglyph = int(t.split(',')[0])
        glyph = self.font_table.glyphs[nglyph]
        glyph_bmp = glyph.pixels
        gidx, blank_prefix, real_char_size, blank_suffix = \
            glyph.width_index, glyph.blank_prefix, \
            glyph.real_char_size, glyph.blank_suffix
        
#        glyph_info, glyph_bmp = self.glyphs[glyph_table][nglyph]

#        gidx, blank_prefix, real_char_size, blank_suffix = glyph_info


        sinfo = '(%d,%d,%d,%d)' % (gidx, blank_prefix, real_char_size, blank_suffix)

        self.frame.debug.WriteText('Loading glyph %d with info %s\n' %
                             (nglyph, sinfo))

        x = 0
#        zoom = 2
        def rep(s,n):
            rs = ''
            for c in s:
                rs += c*n
            return rs

        glyph_bmpx = [glyph_bmp]
        width, height = 16, 16

        for zoom in range(2,4):
            glyph_bmpzx = ''
            
            t = rep(glyph_bmp, zoom)
            for j in range(height):
                glyph_bmpzx += t[j*width*zoom:(j+1)*width*zoom]*zoom
            glyph_bmpx.append(glyph_bmpzx)
            
        for pal in self.font_pals:
            glyph_bmp_paled = ''
            h = 0
            for i in range(len(glyph_bmpx)):
                bmp = wx.BitmapFromBuffer(width*(i+1), height*(i+1),
                                          drawing.indexed_to_rgb(glyph_bmpx[i], pal),
                                          drawing.indexed_to_alpha(glyph_bmpx[i]))
                dc.DrawBitmap(bmp, x*(i+1), h, True)
                h += height*(i+1)
            x += glyph.width#self.font_width_table[gidx]

    def original_mestable_paint(self,evt):
        dc, sel, id = self.mestable_init_dc(evt)
        if id == None:
            return
        t = self.local_db.get_table(sel)
        row = t.get_row(t.select_field('id', id)[0])
        text = row['original']
        if text != None:
            self.mestable_paint(dc, text)

    def on_stringinmestable_select(self,evt):
        sel = self.mestable_list.GetStringSelection().replace('*', '')
        n = int(self.string_mestable_list.GetStringSelection().replace('*', '').split(',')[0])

        t = self.local_db.get_table(sel)

        self.frame.debug.WriteText('Loading row with id %d into table %s\n' %
                             (n, sel))
        row = t.get_row(t.select_field('id', n)[0])

        original_text = lookup(row['original']).replace('[carriage_return]', '\n')
        self.original_text_mestable.ChangeValue(original_text.decode('utf-8'))

        if row['repl'] != None:
            repl_text = lookup(row['repl']).replace('[carriage_return]', '\n')
            self.local_text_mestable.ChangeValue(repl_text.decode('utf-8'))
        else:
            self.local_text_mestable.Clear()

        if row['server'] != None:
            repl_text = lookup(row['server']).replace('[carriage_return]', '\n')
            self.server_text_mestable.ChangeValue(repl_text.decode('utf-8'))
            self.upload_info_mestable.SetLabel('Uploaded by '+str(row['server_uploader'])+' on '+str(row['server_time']))
        else:
            self.upload_info_mestable.SetLabel('No info - fetch table')
            self.server_text_mestable.Clear()

        self.preview.refresh()

  
    def on_mestable_select(self,evt):
        sel = self.mestable_list.GetStringSelection().replace('*', '')
        t = self.local_db.get_table(sel)

        self.frame.debug.WriteText('Selecting table %s (%d rows)\n' %
                             (sel,t.nrows()))

        self.string_mestable_list.Clear()
        self.string_mestable_list.InsertItems(map(lambda n: '%d' % n, range(t.nrows())),
                                              0)

    def load_db(self, debug):
        self.local_db = database.Database()
        self.local_db_changed = False

        mes_res_idx = self.archive.get_file_named('system/rom/JD_message_res_rom.idx')
        mes_res_pak = self.archive.get_file_named('system/rom/JD_message_res.pak')
        mes_res_idxpak = IdxAndPak(mes_res_idx, mes_res_pak)

        self.extra_kanji_tables = {}

        curt = ''

        for tablename in mestables:
            debug.WriteText('Loading table %s.\n' % tablename)
            debug.Update()
            id = mestables[tablename][0]

            if id not in big_text_tables:
#                debug.WriteText('Loading extra kanjis.\n')
#                debug.Update()
                f = mes_res_idxpak.get_file(id)
                if f == None:
                    self.extra_kanji_tables[id] = None
                else:
                    self.extra_kanji_tables[id] = FontTable(f, self.font_width_table)

            if not os.path.exists(self.settings.tables_dir+os.sep+'table%04d.csv' % id):
                debug.WriteText('Table does not exist in database. Dumping it.\n')
                debug.Update()
                ftable = file(self.settings.tables_dir+os.sep+\
                                  'table%04d.csv' % id, 'w')

                if id in big_text_tables:
                    pakname = big_text_tables[id]
                    pak = IdxAndPak(self.archive.get_file_named(idxpak[pakname][0]),
                                    self.archive.get_file_named(idxpak[pakname][1]))

                    bigtext = get_big_text_pack(pak, self.font_width_table)

#                    debug.WriteText('Saving small kanji tables associated to the table.\n')
#                    debug.Update()
                    for j in range(len(bigtext)):
                        if bigtext[j][1] is not None:
                            ffont = file(self.settings.font_dir+os.sep+'extra'+\
                                             os.sep+'font_%02d_%04d.table' % (id,j), 'wb')
                            bigtext[j][1].serialize(ffont)
                            ffont.close()

                    create_table_from_text_list(ftable,
                                                map(lambda c: c[0], bigtext))

                else:             
                    f = self.mespak.get_file(id)
                    nentries = unpack('<H', f.read(2))[0]
                    entries = []
                    for i in range(nentries):
                        entries.append(unpack('<IIH', f.read(10)))
                    t = database.Table(['id', 'original', 'repl', 'repl_time', 'server',
                                        'server_uploader', 'server_time'])
                    for i in range(nentries):
                        slen, soff, boxsize = entries[i]
                        f.seek(soff)
                        t.add_row([i, f.read(slen), None, '', None, '', ''])
                    f.close()
                    t.save(ftable)

                ftable.close()

                ftable = file(self.settings.tables_dir+os.sep+\
                                  'table%04d.csv' % id, 'r')
                t = database.Table(['id', 'original', 'repl', 'repl_time', 'server', 'server_uploader', 'server_time'])
                t.load(ftable)
                ftable.close()
                ftable = file(self.settings.tables_dir+os.sep+\
                                  'trans_table%04d.csv' % id, 'r')
                t_trans = database.Table(['id', 'repl', 'server_uploader', 'server_time'])
                t_trans.load(ftable)
                ftable.close()
                for i in range(t_trans.nrows()):
                    row = t_trans.get_row(i)
                    tid = t.select_field('id', row['id'])[0]
                    t.update_field(tid, 'repl', row['repl'])
                    t.update_field(tid, 'server', row['repl'])
                    t.update_field(tid, 'server_uploader', row['server_uploader'])
                    t.update_field(tid, 'server_time', row['server_time'])

                ftable = file(self.settings.tables_dir+os.sep+\
                                  'table%04d.csv' % id, 'w')
                t.save(ftable)
                ftable.close()

#            debug.WriteText('Populating database table with local csv.\n')
#            debug.Update()
            f = file(self.settings.tables_dir+os.sep+'table%04d.csv' % id, 'r')
            t = database.Table(['id', 'original', 'repl', 'repl_time', 'server', 'server_uploader', 'server_time'])
            t.load(f)

            for i in range(t.nrows()):
                row = t.get_row(i)
                if curt < row['server_time']:
                    curt = row['server_time']

            f.close()

#            debug.WriteText('Loaded table %s (%d rows) with id %d\n' %
#                            (tablename, t.nrows(), id))
#            debug.Update()

            self.local_db.add_table(tablename, t)

        
        self.settings.server_last_update = max(curt, self.settings.server_last_update)
        self.settings.save('editor.cfg')

    def on_idxpak_select(self,evt):
        self.file_in_idxpak_list.Clear()
        k = self.idxpak_list.GetStringSelection()
        idx_fn, pak_fn = idxpak[k]
        self.ipak = IdxAndPak(self.archive.get_file_named(idx_fn),
                              self.archive.get_file_named(pak_fn))
        items = []
        for i in range(len(self.ipak.files)):
            f = self.ipak.files[i]
            if f is not None:
                items.append(str(i))
        self.file_in_idxpak_list.InsertItems(items, 0)

    def on_fileinidx_select(self,evt):
        n = int(self.file_in_idxpak_list.GetStringSelection())
        self.hexdump_idxpak.Clear()
        f = self.ipak.get_file(n)
        self.hexdump_idxpak.WriteText(hexdump_with_table(f).decode('utf-8'))
        f.seek(0)
        self.hexdump_idxpak.SetInsertionPoint(0)
        data = f.read()
        f.seek(0)
        self.data_to_display = data
        self.imgview_idxpak.Refresh()
        self.imgview_idxpak.Update()


    def on_archive_file_select(self,evt):        
        sel = self.frame.archivefileslist.GetStringSelection()
        self.hexdump_pc.Clear()
        splsel = sel.split(',')
        id = int(splsel[0])
        fn = splsel[1]
        f = self.archive.get_file(id,fn=fn)
        self.hexdump_pc.WriteText(hexdump_with_table(f).decode('utf-8'))
        f.seek(0)

        self.hexdump_pc.SetInsertionPoint(0)

    def load_rom(self, romname, debug):
        debug.WriteText('Loading nds rom %s.\n' % romname)
        debug.Update()
        self.nds = NDS(romname)
        pcidx = self.nds.get_file('master/pc.idx')
        pcbin = self.nds.get_file('master/pc.bin')
        debug.WriteText('Building main archive.\n')
        debug.Update()
        self.archive = Archive(pcidx,pcbin)

        self.nc = IdxAndPak(self.archive.get_file_named(idxpak['nc'][0]),
                            self.archive.get_file_named(idxpak['nc'][1]))
        

        self.frame.archivefileslist.Clear()
        filenames = flat_list_by_ids()

        items = []

        sid = -1
        fn = ''
        unsorted = []
        for id in range(len(self.archive.files)):
            fl = self.archive.files[id]
            if fl == None:
                continue

            if type(fl) == type([]):
                for char, pos, f in fl:
                    nfn = ''
                    if id in filenames:
                        for fn in filenames[id]:
                            if len(fn) > pos:
                                if fn[pos] == char:
                                    nfn = fn
                                    break
                    else:
                        nfn = '*'*pos+char
                    unsorted.append(('%d,%s' % (id,nfn), nfn))
            else:
                if id in filenames:
                    unsorted.append(('%d,%s' % (id,filenames[id][0]), 
                                     filenames[id][0]))
                else:
                    unsorted.append(('%d,' % id, ''))

        unsorted.sort(lambda x,y:cmp(x[1],y[1]))
        
        items = map(lambda x:x[0],unsorted)
        
        self.frame.archivefileslist.InsertItems(items,0)

        debug.WriteText('Loading main message pack.\n')
        debug.Update()
        mesidx = self.archive.get_file_named('system/rom/JD_message_rom.idx')
        mespak = self.archive.get_file_named('system/rom/JD_message.pak')
        self.mespak = IdxAndPak(mesidx, mespak)

        debug.WriteText('Loading font.\n')
        debug.Update()

        self.load_font()


    def load_font(self):
        self.font_width_table = []
        font_width = self.archive.get_file_named('menu/font/FontWidthTable.bin')
        nentries = unpack('<H', font_width.read(2))[0]

        for i in range(nentries):
            self.font_width_table.append(ord(font_width.read(1)))

        font_table = self.archive.get_file_named('menu/font/UseMoji_image.bin')

        self.font_table = FontTable(font_table, self.font_width_table, lookup, load_lazy=True)

        mainroot = self.font_tree.AddRoot('main')

        fim = Image.open(self.settings.font_dir+os.sep+'font_new.png')

        self.font_table.load_image(fim)

        for i in range(len(self.font_table.glyphs)):
            sglyph = str(i)+','+self.font_table.glyphs[i].value
            self.font_tree.AppendItem(mainroot, sglyph.decode('utf-8'))

        font_palettes = self.archive.get_file_named('menu/font/fontpalette.bin')
        font_palettes.read(1)
        self.font_pals = []
        for i in range(20):
            pal = []
            for j in range(4):
                pal.append(drawing.unpack_rgb15(unpack('<H',font_palettes.read(2))[0]))

            self.font_pals.append(pal)

    def OnInit(self):
        self.res = app_xrc.get_resources()

        splash = app_xrc.xrcsplash_screen(None)
        splash.version_label.SetLabel('Version %s-%s by %s (released on %s) [%s]' %\
                                          (version, version_type, author, date,
                                           website))

        splash.Show()

        self.local_db = None

        splash.status.WriteText('Loading settings.\n')
        splash.status.Update()
        self.settings = Settings()
        self.settings.load('editor.cfg')

        self.frame = MainFrame(self)

        if self.settings.original_rom == None:
            splash.status.WriteText('Asking for original rom.\n')
            splash.status.Update()

            curd = os.getcwd()
            fdlg = wx.FileDialog(
                self.frame, message="Choose the original rom of FFTA2",
                defaultDir=os.getcwd(), 
                defaultFile="",
                wildcard="NDS rom file (*.nds)|*.nds",
                style=wx.OPEN | wx.CHANGE_DIR
                )
            if fdlg.ShowModal() == wx.ID_OK:
                self.settings.original_rom = fdlg.GetPath()
                self.settings.patched_rom = self.settings.original_rom[:-4]+'p.nds'
                os.chdir(curd)
                fdlg.Destroy()
            else:
                fdlg.Destroy()
                self.frame.Destroy()

        self.load_rom(self.settings.original_rom, debug=splash.status)

#        get_story_pack(self.archive, self.font_width_table, save_table=True)

        self.load_db(debug=splash.status)

        splash.status.WriteText('Load finished\n')
        splash.status.Update()

        splash.status.WriteText('Updating tables\n.')
        splash.status.Update()

        if self.need_update() and self.settings.auto_fetch:
            for tablen in mestables:
                splash.status.WriteText('Fetching table %s.\n' % tablen)
                splash.status.Update()
                mt, changed = self.fetch_table(tablen, copylocal=True, debug=splash.status)
                if mt > self.settings.server_last_update:
                    self.settings.server_last_update = mt
            self.on_save()
            self.settings.save('editor.cfg')

        self.frame.debug.WriteText(splash.status.GetValue())

        splash.Destroy()
        self.frame.Show()
        return True
    
    def on_quit(self, evt):
        if self.local_db_changed:
            dlg = wx.MessageDialog(self.frame,
                                   'The local database has changed\nAre you sure you want to quit',
                                   'Warning',
                                   wx.YES_NO | wx.NO_DEFAULT | wx.ICON_EXCLAMATION)
            res = dlg.ShowModal()
            if res == wx.ID_NO:
                evt.Veto()
                return

        self.settings.save('editor.cfg')
        self.frame.Destroy()

if __name__ == "__main__":
    socket.setdefaulttimeout(timeout)
    app = MainApp(False)
    app.MainLoop()
