import sys
from struct import pack, unpack
import crc
from fileinfile import FileInFile
from tempfile import TemporaryFile

def protect(d):
    if type(d) == str:
        if d.isalnum():
            return d
        else:
            s = '0x'
            for c in d:
                s += hex(ord(c))[2:]
            return s
    else:
        return hex(d)

class FatFile(object):
    def __init__(self, name, f):
        self.name = name
        self.f = f

    def __str__(self,prefix=''):
        return prefix+'/'+self.name

class FatDir(object):
    def __init__(self,name=''):
        self.files = []
        self.sub_dirs = []
        self.name = name
        self.overlays = []

    def search(self, name):
        r = name.split('/')
        if r[0] == self.name:
            rn = '/'.join(r[1:])
            for f in self.files:
                if rn == f.name:
                    return f
            for d in self.sub_dirs:
                rt = d.search(rn)
                if rt != None:
                    return rt
        return None

    def populate(self, f, dir_id, fto, fts, fo, fs, noverlay=None,
                 dontloadfiles=False):
        f.seek(fto+8*(dir_id & 0xFFF))
        entry_start, top_file_id, parent_id = unpack('<IHH', f.read(8))
        i = top_file_id

        if noverlay != None:
            f.seek(fo)
            for k in range(noverlay):
                spos = f.tell()
                f.seek(fo+8*k)
                top, bottom = unpack('<II', f.read(8))
                f.seek(top)
                self.overlays += [FatFile('#%04d' % k, 
                                          FileInFile(f,top,bottom-top))]
        f.seek(fto+entry_start)

        while True:
            if f.tell() >= fto+fts:
                break
            elen = ord(f.read(1))
            edir = elen & 128
            nlen = elen & 127
            if nlen == 0:
                break
            name = f.read(nlen)

            if edir:
                sub_dir_id = unpack('<H', f.read(2))[0]
                spos = f.tell()
                d = FatDir(name)
                #print sub_dir_id & 0xfff, name
                d.populate(f, sub_dir_id, fto, 
                           fts, fo, fs)
                self.sub_dirs += [d]
                f.seek(spos)
            else:
                spos = f.tell()
                f.seek(fo+8*i)
                top, bottom = unpack('<II', f.read(8))
                f.seek(top)
                #print i, name, top
                if dontloadfiles:
                    self.files += [FatFile(name, '', top)]
                else:
                    self.files += [FatFile(name, FileInFile(f,top,bottom-top))]
                f.seek(spos)
                i += 1

    def width(self):
        n = 1
        for d in self.sub_dirs:
            n += d.width()
        return n

    def table_size(self):
        n = 8+len(self.name)+2
        for d in self.sub_dirs:
            n += d.table_size()+2
        for f in self.files:
            n += len(f.name)+1
        return n

    def serialize(self,dir_id=0xF000,parent_id=None,name_off=0,cur_id=0,fat_table_offset=0):
        if parent_id == None:
            nd = self.width()
            cur_name_off = nd*8
            cur_id = len(self.overlays)
            s = pack('<IHH', cur_name_off, cur_id, nd)
        else:
            cur_name_off = name_off
            s = pack('<IHH', cur_name_off, cur_id, parent_id)
        sn = ''
        sub_dir_id = dir_id + 1
        file_id = cur_id+len(self.files)

        id_files = []
        for i in range(len(self.files)):
            id_files += [(cur_id+i, self.files[i])]

        sub_ids = {}

        def my_cmp(a,b):
            return a.name < b.name

        subd = self.sub_dirs
        subf = self.files

        subf.sort(my_cmp)
        subd.sort(my_cmp)

        sub = subf+subd
        for e in sub:
            ln = len(e.name)
            if type(e) == FatDir:
                ln |= 0x80
            sn += chr(ln)+e.name
            if type(e) == FatDir:
                sn += pack('<H', sub_dir_id)
                sub_ids[e] = sub_dir_id
                sub_dir_id += e.width()
        sn += '\x00'
        cur_name_off += len(sn)

        for d in self.sub_dirs:
            e,n,id,l = d.serialize(sub_ids[d],dir_id,cur_name_off,file_id)
            file_id = id
            id_files += l
            cur_name_off += len(n)
            s += e
            sn += n
            
        if parent_id == None:
            fat = ''
            fat_table = s+sn
            excur = fat_table_offset+len(fat_table)
            fat_off = ((excur >> 8) + 1)<<8
            fat = fat_table+'\x00'*(fat_off-excur)
            size_table = file_id*8
            excur = fat_off + size_table
            ban_off = fat_off + (((size_table >> 8) + 1) << 8)

            cur_off = ban_off + 0x840
            cur_off = (cur_off + 0x1ff) & ~0x1ff

            table = ''
            sf = '\x00'*(cur_off-excur)
            id_files = [(i, self.overlays[i]) for i in range(len(self.overlays))] + id_files

            fatfile = TemporaryFile()
            fatfile.write(sf)
            i = 0
            for id,f in id_files:
                f.f.seek(0,2)
                fsize = f.f.tell()
                f.f.seek(0)

                #print 'Repacking %s file (%d bytes) %s' % \
                #    (f.name, fsize, repr(f.f))

                table += pack('<II', cur_off, cur_off+fsize)

#                fdata = f.f.read()
                
#                if f.name == 'pc.idx':
#                    n = unpack('<H', f.f.read(2))[0]
#                    print n
#                    f.f.seek(0)
                fatfile.write(f.f.read())

                cur_off += fsize

                excur = cur_off
                cur_off = (cur_off + 0x1ff) & ~0x1ff
                i += 1
                if i != len(id_files):
                    fatfile.write('\x00'*(cur_off-excur))


            #fat+= table+sf
            fullfat = TemporaryFile()
            fullfat.write(fat)
            fullfat.write(table)
            fatfile.seek(0)
            fullfat.write(fatfile.read())
            fatfile.close()
            fullfat.seek(0)

            return fullfat, fat_off, ban_off, len(fat_table), size_table
        else:
            return s, sn, file_id, id_files

    def __str__(self,prefix=''):
        nprefix = prefix+'/'+self.name
        s = ''
        for d in self.sub_dirs:
            s += d.__str__(nprefix)
        for f in self.files:
            s += f.__str__(nprefix)
        return s

class NDS(object):
    def __init__(self, fn, dontloadfiles=False):
        self.pack = [
            ('title', 'S', 12),
            ('code', 'S', 4),
            ('maker_code', 'S', 2),
            ('unit_code', 'S', 1),
            ('device_code', 'S', 1),
            ('card_size', 'B'),
            ('card_info', 'S', 10),
            ('flags', 'B'),
            ('arm9_source', 'I'),
            ('arm9_exec_addr', 'I'),
            ('arm9_copy_addr', 'I'),
            ('arm9_size', 'I'),
            ('arm7_source', 'I'),
            ('arm7_exec_addr', 'I'),
            ('arm7_copy_addr', 'I'),
            ('arm7_size', 'I'),
            ('fat_table_offset', 'I'),
            ('fat_table_size', 'I'),
            ('fat_offset', 'I'),
            ('fat_size', 'I'),
            ('arm9_overlay', 'I'),
            ('arm9_overlay_size', 'I'),
            ('arm7_overlay', 'I'),
            ('arm7_overlay_size', 'I'),
            ('control_read_flags', 'S', 4),
            ('control_init_flags', 'S', 4),
            ('icon_titles', 'I'),
            ('crc16', 'H'),
            ('rom_timeout', 'H'),
            ('arm9_unk_addr', 'I'),
            ('arm7_unk_addr', 'I'),
            ('magic', 'S', 8),
            ('rom_size', 'I'),
            ('header_size', 'I'),
            ('unknown', 'S', 56),
            ('gba_logo', 'S', 156),
            ('logo_crc16', 'H'),
            ('header_crc16', 'H'),
            ('reserved', 'S', 160)
            ]

        self.fn = fn
        self.f = file(fn,'rb')
        self.header = {}
        self.unpack()

        self.f.seek(self.header['icon_titles'])
        self.banneer = self.f.read(0x840)

        self.f.seek(self.header['arm9_source'])
        self.arm9 = self.f.read(self.header['arm9_size'])
        self.f.seek(self.header['arm7_source'])
        self.arm7 = self.f.read(self.header['arm7_size'])
        self.f.seek(self.header['arm9_overlay'])
        self.arm9_overlay = self.f.read(self.header['arm9_overlay_size'])

        self.overlay_table = []
        for i in range(self.header['arm9_overlay_size']/32):
            self.overlay_table.append(unpack('<IIIIIIII', 
                                             self.arm9_overlay[32*i:32*(i+1)]))

        self.f.seek(self.header['arm7_overlay'])
        self.arm7_overlay = self.f.read(self.header['arm7_overlay_size'])

        #print hex(self.header['arm9_overlay_size']), hex(self.header['arm7_overlay_size'])

        for i in range(len(self.overlay_table)):
            overlay_id, ram_address, ram_size, bss_size, \
                static_init_start, static_init_end, file_id, reserved = self.overlay_table[i]
         #   print map(hex, self.overlay_table[i])

        self.fat = FatDir('root')
        self.fat.populate(self.f, 0xF000,
                          self.header['fat_table_offset'], 
                          self.header['fat_table_size'],
                          self.header['fat_offset'], 
                          self.header['fat_table_size'],
                          self.header['arm9_overlay_size']/32,
                          dontloadfiles=dontloadfiles)

        for pack_desc in self.pack:
            pname = pack_desc[0]
            ptype = pack_desc[1]
       #     if ptype !='S':
       #         print pname+' = '+hex(self.header[pname])
       # print hex(self.header['logo_crc16'])

    def close(self):
        self.f.close()

    def unpack(self):
        for pack_desc in self.pack:
            pname = pack_desc[0]
            ptype = pack_desc[1]
            if ptype == 'S':
                self.header[pname] = self.f.read(int(pack_desc[2]))
            elif ptype == 'I':
                self.header[pname] = unpack('<I', self.f.read(4))[0]
            elif ptype == 'H':
                self.header[pname]= unpack('<H', self.f.read(2))[0]
            elif ptype == 'B':
                self.header[pname] = unpack('<B', self.f.read(1))[0]

    def serialize_header(self):
        s = ''
        for pack_desc in self.pack:
            pname = pack_desc[0]
            ptype = pack_desc[1]
            if ptype == 'S':
                s += self.header[pname]
            else:
                s += pack('<'+ptype, self.header[pname])
        return s

    def get_file(self, fn):
        return self.fat.search('root/'+fn).f

    def replace_file(self, fn, f):
        self.fat.search('root/'+fn).f = f

    def save(self, nfn):
        def align(n):
            return (n + 0x1ff) & ~0x1ff
 
        f = file(nfn, 'wb')

        size_header = 0x200
        arm9_off = 0x4000

        arm9_size = len(self.arm9)
        f.seek(arm9_off)
        f.write(self.arm9)
        f.write(pack('<III', 0xDEC00621, 0xAD8, 0))

        arm9_overlay_size = len(self.arm9_overlay)

        if arm9_overlay_size:
            arm9_overlay_off = arm9_off + arm9_size + 12
            f.write(self.arm9_overlay)
            arm7_off = align(arm9_overlay_off+arm9_overlay_size)
        else:
            arm7_off = align(arm9_off+arm9_size+12)
            arm9_overlay_off = 0

        curoff = f.tell()
        f.write('\xff'*(arm7_off-curoff))

        f.write(self.arm7)

        arm7_size = len(self.arm7)

        arm7_overlay_size = len(self.arm7_overlay)

        if arm7_overlay_size:
            arm7_overlay_off = arm7_off + arm7_size
            fat_table_off = align(arm7_overlay_off+arm7_overlay_size)
        else:
            fat_table_off = align(arm7_off+arm7_size)
            arm7_overlay_off = 0

        f.seek(fat_table_off)
        fat, fat_offset, bann_offset, sft, sf = self.fat.serialize(fat_table_offset=fat_table_off)
        f.write(fat.read())
        fat.close()

        rom_size = f.tell()

        f.seek(bann_offset)
        f.write(self.banneer)

        self.header['fat_offset'] = fat_offset
        self.header['fat_table_offset'] = fat_table_off
        self.header['icon_titles'] = bann_offset
        self.header['arm9_source'] = arm9_off
        self.header['arm9_size'] = arm9_size
        self.header['arm7_source'] = arm7_off
        self.header['arm7_size'] = arm7_size
        self.header['arm9_overlay'] = arm9_overlay_off
        self.header['arm9_overlay_size'] = arm9_overlay_size
        self.header['arm7_overlay'] = arm7_overlay_off
        self.header['arm7_overlay_size'] = arm7_overlay_size
        self.header['fat_size'] = sf
        self.header['fat_table_size'] = sft
        self.header['rom_size'] = rom_size

        head = self.serialize_header()
        self.header['logo_crc16'] = crc.calc_crc16(head[0xc0:0x15c])

        head = self.serialize_header()
        self.header['header_crc16'] = crc.calc_crc16(head[:0x15e])

        f.seek(0)
        f.write(self.serialize_header())

        f.close()

if __name__=="__main__":
    nds = NDS('ffa2.nds')
#    nds.save('ffa2p.nds')
    nds.close()

