import wx
import os
from drawing import draw_text
from table import reverse
from font import FontTable

class Preview(object):
    def __init__(self, frame):
        self.frame = frame
        self.preview_modes = [
            ('default', None, [(0,(0,0),0)]),
            ('jobs/abilities descriptions', 'preview_abilities.png',
             [(0, (7,11), 16), (3-31, (18,51), 16)]),
            ('sub area description', 'preview_subarea.png',
             [(0,(37,15),10), (4-32,(62,0),13)]),
            ('area description', 'preview_area.png', [(0,(6,6),11)]),
            ('red box message', 'preview_redbox.png', [(0,(8,13),0)]),
            ('big text', 'preview_bigtext.png', [(0,(7,21),0)]),
            ('character talk', 'preview_talk.png', [(0,(18,26),10)]),
            ]
        self.cached_images = {}
        for mode in self.preview_modes:
            imn = mode[1]
            if imn is None:
                continue
            if imn not in self.cached_images:
                self.cached_images[imn] = wx.Image(self.frame.parent_app.settings.preview_dir+os.sep+imn)

        self.mode = 0

        self.frame.preview_window.Bind(wx.EVT_PAINT, self.paint)

        l = []
        for mode in self.preview_modes:
            l += [mode[0]]
        l.reverse()

        self.frame.preview_choice.Clear()
        for e in l:
            self.frame.preview_choice.Insert(e,0)
        self.frame.preview_choice.SetSelection(0)
        
        self.frame.preview_choice.Bind(wx.EVT_CHOICE, self.change_preview_mode)

    def paint(self, evt):
        dc = wx.PaintDC(self.frame.preview_window)
        self.frame.preview_window.PrepareDC(dc)
        dc.Clear()
        sel = self.frame.mestable_list.GetStringSelection().replace('*', '')
        if sel == '':
            return
        s = self.frame.string_mestable_list.GetStringSelection().replace('*','').split(',')[0]
        if s == '':
            return
        id = int(s)

        try:
            mode = self.preview_modes[self.mode]
        except IndexError:
            return

        tableid = 0
        for k in self.frame.mestables:
            if self.frame.mestables[k] == sel:
                tableid = k
                break
        if mode[1] is None:
            image = None
        else:
            image = self.cached_images[mode[1]]
        self.draw(dc, tableid, id, image, mode[2])

    def draw(self, dc, tablen, strid, image, pos_params):
        if image is not None:
            w, h = image.GetWidth(), image.GetHeight()
        else:
            w = 200
            h = 100

        self.frame.preview_window.SetVirtualSize((w*3,h))

        l = ['repl', 'original', 'server']
        i = 0
        for f in l:
            if image is not None:
                bmp = wx.BitmapFromImage(image)
                dc.DrawBitmap(bmp, w*i, 0, True)

            for toff, pos, paln in pos_params:
                extra_table = None
                if tablen+toff in self.frame.mestables:
                    t = self.frame.parent_app.local_db.get_table(self.frame.mestables[tablen+toff])

                    strselect = t.select_field('id', strid)

                    if tablen+toff in self.frame.parent_app.extra_kanji_tables:
                        extra_table = self.frame.parent_app.extra_kanji_tables[tablen+toff]
                    else:
                        font_dir = self.frame.parent_app.settings.font_dir
                        font_extra_file = font_dir + os.sep + 'extra' + os.sep +\
                            'font_%02d_%04d.table' % (tablen+toff, strid)

                        if os.path.exists(font_extra_file):
                            fontf = file(font_extra_file, 'rb')
                            extra_table = FontTable(fontf,
                                                    self.frame.parent_app.font_width_table)
                            fontf.close()
                        
                    if len(strselect) == 0:
                        continue

                    text = t.get_field(strselect[0], f)

                    if text == None:
                        continue
                else:
                    text = reverse('Invalid')

                draw_text(dc, text, pos[0]+w*i, pos[1], self.frame.parent_app.font_width_table,
                          self.frame.parent_app.font_table, extra_table,
                          self.frame.parent_app.font_pals[paln])
            i += 1


    def refresh(self):
        self.frame.preview_window.Refresh()
        self.frame.preview_window.Update()

    def change_preview_mode(self, evt):
        self.mode = self.frame.preview_choice.GetSelection()
        self.refresh()
    
